use std::rc::Rc;

use crate::{
    Builtin, DeBrujinAst,
    exec::builtins::DeBrujinBuiltInAst,
    parse::Constant,
    types::{PrimitiveType, TaggedType, Type, TypeTag},
};

pub struct AddOp;
struct AddOpNat(usize);
struct AddOpFloat(f64);

impl Builtin for AddOp {
    fn name(&self) -> String {
        "add".to_string()
    }

    fn r#type(&self) -> TaggedType {
        TaggedType::Tagged(
            TypeTag::Num,
            "a".to_string(),
            Box::new(TaggedType::Concrete(Type::arrow(
                "a",
                Type::arrow("a", "a"),
            ))),
        )
    }

    fn apply(&self, rhs: DeBrujinBuiltInAst) -> Option<DeBrujinBuiltInAst> {
        match rhs {
            DeBrujinBuiltInAst::Constant(Constant::Nat(n)) => {
                Some(DeBrujinBuiltInAst::Builtin(Rc::new(AddOpNat(n))))
            }
            DeBrujinBuiltInAst::Constant(Constant::Float(n)) => {
                Some(DeBrujinBuiltInAst::Builtin(Rc::new(AddOpFloat(n))))
            }
            _ => None,
        }
    }
}

impl Builtin for AddOpNat {
    fn name(&self) -> String {
        format!("add{}", self.0)
    }

    fn to_ast(&self) -> DeBrujinAst {
        DeBrujinAst::Application(
            Box::new(DeBrujinAst::FreeVariable("add".to_string())),
            Box::new(DeBrujinAst::Constant(Constant::Nat(self.0))),
        )
    }

    fn r#type(&self) -> TaggedType {
        Type::arrow(PrimitiveType::Nat, PrimitiveType::Nat).into()
    }

    fn apply(&self, rhs: DeBrujinBuiltInAst) -> Option<DeBrujinBuiltInAst> {
        match rhs {
            DeBrujinBuiltInAst::Constant(Constant::Nat(n)) => {
                Some(DeBrujinBuiltInAst::Constant(Constant::Nat(n + self.0)))
            }
            _ => None,
        }
    }
}

impl Builtin for AddOpFloat {
    fn name(&self) -> String {
        format!("add{}", self.0)
    }

    fn to_ast(&self) -> DeBrujinAst {
        DeBrujinAst::Application(
            Box::new(DeBrujinAst::FreeVariable("add".to_string())),
            Box::new(DeBrujinAst::Constant(Constant::Float(self.0))),
        )
    }

    fn r#type(&self) -> TaggedType {
        Type::arrow(PrimitiveType::Float, PrimitiveType::Float).into()
    }

    fn apply(&self, rhs: DeBrujinBuiltInAst) -> Option<DeBrujinBuiltInAst> {
        match rhs {
            DeBrujinBuiltInAst::Constant(Constant::Float(n)) => {
                Some(DeBrujinBuiltInAst::Constant(Constant::Float(n + self.0)))
            }
            _ => None,
        }
    }
}

pub struct SubOp;
struct SubOpNat(usize);
struct SubOpFloat(f64);

impl Builtin for SubOp {
    fn name(&self) -> String {
        "sub".to_string()
    }

    fn r#type(&self) -> TaggedType {
        TaggedType::Tagged(
            TypeTag::Num,
            "a".to_string(),
            Box::new(TaggedType::Concrete(Type::arrow(
                "a",
                Type::arrow("a", "a"),
            ))),
        )
    }

    fn apply(&self, rhs: DeBrujinBuiltInAst) -> Option<DeBrujinBuiltInAst> {
        match rhs {
            DeBrujinBuiltInAst::Constant(Constant::Nat(n)) => {
                Some(DeBrujinBuiltInAst::Builtin(Rc::new(SubOpNat(n))))
            }
            DeBrujinBuiltInAst::Constant(Constant::Float(n)) => {
                Some(DeBrujinBuiltInAst::Builtin(Rc::new(SubOpFloat(n))))
            }
            _ => None,
        }
    }
}

impl Builtin for SubOpNat {
    fn name(&self) -> String {
        format!("sub{}", self.0)
    }

    fn to_ast(&self) -> DeBrujinAst {
        DeBrujinAst::Application(
            Box::new(DeBrujinAst::FreeVariable("sub".to_string())),
            Box::new(DeBrujinAst::Constant(Constant::Nat(self.0))),
        )
    }

    fn r#type(&self) -> TaggedType {
        Type::arrow(PrimitiveType::Nat, PrimitiveType::Nat).into()
    }

    fn apply(&self, rhs: DeBrujinBuiltInAst) -> Option<DeBrujinBuiltInAst> {
        match rhs {
            DeBrujinBuiltInAst::Constant(Constant::Nat(n)) => {
                Some(DeBrujinBuiltInAst::Constant(Constant::Nat(self.0 - n)))
            }
            _ => None,
        }
    }
}

impl Builtin for SubOpFloat {
    fn name(&self) -> String {
        format!("sub{}", self.0)
    }

    fn to_ast(&self) -> DeBrujinAst {
        DeBrujinAst::Application(
            Box::new(DeBrujinAst::FreeVariable("sub".to_string())),
            Box::new(DeBrujinAst::Constant(Constant::Float(self.0))),
        )
    }

    fn r#type(&self) -> TaggedType {
        Type::arrow(PrimitiveType::Float, PrimitiveType::Float).into()
    }

    fn apply(&self, rhs: DeBrujinBuiltInAst) -> Option<DeBrujinBuiltInAst> {
        match rhs {
            DeBrujinBuiltInAst::Constant(Constant::Float(n)) => {
                Some(DeBrujinBuiltInAst::Constant(Constant::Float(self.0 - n)))
            }
            _ => None,
        }
    }
}

pub struct MulOp;
struct MulOpNat(usize);
struct MulOpFloat(f64);

impl Builtin for MulOp {
    fn name(&self) -> String {
        "mul".to_string()
    }

    fn r#type(&self) -> TaggedType {
        TaggedType::Tagged(
            TypeTag::Num,
            "a".to_string(),
            Box::new(TaggedType::Concrete(Type::arrow(
                "a",
                Type::arrow("a", "a"),
            ))),
        )
    }

    fn apply(&self, rhs: DeBrujinBuiltInAst) -> Option<DeBrujinBuiltInAst> {
        match rhs {
            DeBrujinBuiltInAst::Constant(Constant::Nat(n)) => {
                Some(DeBrujinBuiltInAst::Builtin(Rc::new(MulOpNat(n))))
            }
            DeBrujinBuiltInAst::Constant(Constant::Float(n)) => {
                Some(DeBrujinBuiltInAst::Builtin(Rc::new(MulOpFloat(n))))
            }
            _ => None,
        }
    }
}

impl Builtin for MulOpNat {
    fn name(&self) -> String {
        format!("mul{}", self.0)
    }

    fn to_ast(&self) -> DeBrujinAst {
        DeBrujinAst::Application(
            Box::new(DeBrujinAst::FreeVariable("mul".to_string())),
            Box::new(DeBrujinAst::Constant(Constant::Nat(self.0))),
        )
    }

    fn r#type(&self) -> TaggedType {
        Type::arrow(PrimitiveType::Nat, PrimitiveType::Nat).into()
    }

    fn apply(&self, rhs: DeBrujinBuiltInAst) -> Option<DeBrujinBuiltInAst> {
        match rhs {
            DeBrujinBuiltInAst::Constant(Constant::Nat(n)) => {
                Some(DeBrujinBuiltInAst::Constant(Constant::Nat(n * self.0)))
            }
            _ => None,
        }
    }
}

impl Builtin for MulOpFloat {
    fn name(&self) -> String {
        format!("mul{}", self.0)
    }

    fn to_ast(&self) -> DeBrujinAst {
        DeBrujinAst::Application(
            Box::new(DeBrujinAst::FreeVariable("mul".to_string())),
            Box::new(DeBrujinAst::Constant(Constant::Float(self.0))),
        )
    }

    fn r#type(&self) -> TaggedType {
        Type::arrow(PrimitiveType::Float, PrimitiveType::Float).into()
    }

    fn apply(&self, rhs: DeBrujinBuiltInAst) -> Option<DeBrujinBuiltInAst> {
        match rhs {
            DeBrujinBuiltInAst::Constant(Constant::Float(n)) => {
                Some(DeBrujinBuiltInAst::Constant(Constant::Float(n * self.0)))
            }
            _ => None,
        }
    }
}
