pub mod builtin_definitions;
mod builtins;
#[cfg(test)]
mod test;

pub use builtin_definitions as builtin;
pub use builtins::Builtin;

use std::{collections::HashMap, rc::Rc};

use crate::{
    parse::{Ast, Constant, Ident},
    types::TaggedType,
    vec_map::VecMap,
};

impl From<Ast> for DeBrujinAst {
    fn from(val: Ast) -> Self {
        val.into_de_brujin_inter(Rc::new(HashMap::new()))
    }
}

impl Ast {
    fn into_de_brujin_inter(self, mut gamma: Rc<HashMap<String, usize>>) -> DeBrujinAst {
        match self {
            Ast::Abstraction(var, t, ast) => {
                let gamma_ref = Rc::make_mut(&mut gamma);
                gamma_ref.values_mut().for_each(|v| *v += 1);
                gamma_ref.insert(var.clone(), 1);

                DeBrujinAst::Abstraction(var, t, Box::new(ast.into_de_brujin_inter(gamma)))
            }
            Ast::Application(lhs, rhs) => DeBrujinAst::Application(
                Box::new(lhs.into_de_brujin_inter(gamma.clone())),
                Box::new(rhs.into_de_brujin_inter(gamma)),
            ),
            Ast::Variable(v) => {
                if let Some(c) = gamma.get(&v) {
                    DeBrujinAst::BoundVariable(*c)
                } else {
                    DeBrujinAst::FreeVariable(v)
                }
            }
            Ast::Constant(constant) => DeBrujinAst::Constant(constant),
        }
    }
}

#[derive(Debug, Clone, PartialEq)]
pub enum DeBrujinAst {
    Abstraction(Ident, Option<TaggedType>, Box<DeBrujinAst>), // \:1.2
    Application(Box<DeBrujinAst>, Box<DeBrujinAst>),          // 0 1
    FreeVariable(String),                                     // x
    BoundVariable(usize),                                     // 1
    Constant(Constant),                                       // true | false | n
}

impl From<DeBrujinAst> for Ast {
    fn from(val: DeBrujinAst) -> Self {
        val.into_ast(Rc::new(VecMap::new()))
    }
}

impl DeBrujinAst {
    pub fn reduce(self, builtins: &HashMap<String, Rc<dyn Builtin>>) -> DeBrujinAst {
        let ast = self.resolve_builtins(builtins);
        let ast = ast.reduce();
        ast.into()
    }

    // Vec<(usize, String)> as opposed to a HashMap here since we need to mutate
    // all the keys every time recuse into an Abstraction, which would be suuper
    // expensive if not impossible with a HashMap
    fn into_ast(self, mut gamma: Rc<VecMap<usize, String>>) -> Ast {
        match self {
            DeBrujinAst::Abstraction(i, t, de_brujin_ast) => {
                let gamma_ref = Rc::make_mut(&mut gamma);
                gamma_ref.map_keys(|i| *i += 1);
                gamma_ref.insert(1, i.clone());

                Ast::Abstraction(i, t, Box::new(de_brujin_ast.into_ast(gamma)))
            }
            DeBrujinAst::Application(lhs, rhs) => Ast::Application(
                Box::new(lhs.into_ast(gamma.clone())),
                Box::new(rhs.into_ast(gamma)),
            ),
            DeBrujinAst::FreeVariable(i) => Ast::Variable(i),
            DeBrujinAst::BoundVariable(n) => Ast::Variable(
                gamma
                    .get(&n)
                    .unwrap() // Compiler bug if panics
                    .clone(),
            ),
            DeBrujinAst::Constant(constant) => Ast::Constant(constant),
        }
    }
}
