use std::{fmt::Debug, mem, vec};

pub struct MultiMap<K: Eq, V> {
    map: Vec<(K, V)>,
}

impl<K: Eq, V> MultiMap<K, V> {
    pub fn new() -> Self {
        Self { map: Vec::new() }
    }

    pub fn insert(&mut self, key: K, val: V) {
        self.map.push((key, val));
    }

    pub fn get(&self, key: &K) -> Vec<&V> {
        self.map
            .iter()
            .filter_map(|(k, v)| (k == key).then_some(v))
            .collect()
    }

    pub fn map_keys<F: Fn(&mut K)>(&mut self, f: F) {
        self.map.iter_mut().for_each(|(k, _)| f(k));
    }

    pub fn map<F: Fn((K, V)) -> (K, V)>(&mut self, f: F) {
        let vec = mem::take(&mut self.map);
        let vec = vec.into_iter().map(f).collect::<Vec<(K, V)>>();
        self.map = vec;
    }

    pub fn try_map<E, F: Fn((K, V)) -> Result<(K, V), E>>(&mut self, f: F) -> Result<(), E> {
        let vec = mem::take(&mut self.map);
        let vec = vec.into_iter().map(f).collect::<Result<Vec<(K, V)>, _>>()?;
        self.map = vec;
        Ok(())
    }

    pub fn find<F: FnMut(&&(K, V)) -> bool>(&self, f: F) -> Option<&(K, V)> {
        self.map.iter().find(f)
    }

    pub fn find_remove<F: FnMut(&&(K, V)) -> bool>(&mut self, mut f: F) -> Option<(K, V)> {
        let idx = self
            .map
            .iter()
            .enumerate()
            .find_map(|(n, ref kv)| f(kv).then_some(n))?;
        Some(self.map.swap_remove(idx))
    }

    pub fn len(&self) -> usize {
        self.map.len()
    }

    pub fn is_empty(&self) -> bool {
        self.map.is_empty()
    }

    pub fn pop(&mut self) -> Option<(K, V)> {
        self.map.pop()
    }
}

impl<K: Eq, V> IntoIterator for MultiMap<K, V> {
    type Item = (K, V);

    type IntoIter = vec::IntoIter<(K, V)>;

    fn into_iter(self) -> Self::IntoIter {
        self.map.into_iter()
    }
}

impl<K: Eq, V> FromIterator<(K, V)> for MultiMap<K, V> {
    fn from_iter<T: IntoIterator<Item = (K, V)>>(iter: T) -> Self {
        Self {
            map: iter.into_iter().collect(),
        }
    }
}

impl<K: Eq, V> Default for MultiMap<K, V> {
    fn default() -> Self {
        Self {
            map: Vec::default(),
        }
    }
}

impl<K: Eq + Clone, V: Clone> Clone for MultiMap<K, V> {
    fn clone(&self) -> Self {
        Self {
            map: self.map.clone(),
        }
    }
}

impl<K: Eq + Debug, V: Debug> Debug for MultiMap<K, V> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_map()
            .entries(self.map.iter().map(|(k, v)| (k, v)))
            .finish()
    }
}
