use crate::{
    parse::{Ast, Constant, tokenize::Lexer},
    types::{PrimitiveType, TaggedType, Type, TypeTag},
};

use super::parser;

#[test]
fn parse_constant() {
    let parser = parser::AstParser::new();
    let ast = parser.parse(Lexer::new("5")).unwrap();
    assert_eq!(ast, Ast::Constant(Constant::Nat(5)));

    let ast = parser.parse(Lexer::new("3.14")).unwrap();
    assert_eq!(ast, Ast::Constant(Constant::Float(3.14)));

    let ast = parser.parse(Lexer::new("true")).unwrap();
    assert_eq!(ast, Ast::Constant(Constant::Bool(true)));
}

#[test]
fn parse_variable() {
    let parser = parser::AstParser::new();
    let ast = parser.parse(Lexer::new("x")).unwrap();
    assert_eq!(ast, Ast::Variable("x".to_string()));
}

#[test]
fn parse_type() {
    let parser = parser::TaggedTypeParser::new();
    let ast = parser.parse(Lexer::new("Num a => a -> a -> a")).unwrap();
    assert_eq!(
        ast,
        TaggedType::Tagged(
            TypeTag::Num,
            "a".to_string(),
            Box::new(TaggedType::Concrete(Type::arrow(
                "a",
                Type::arrow("a", "a")
            )))
        )
    );

    let ast = parser
        .parse(Lexer::new("Any a => (a -> a) -> a -> a"))
        .unwrap();
    assert_eq!(
        ast,
        TaggedType::Tagged(
            TypeTag::Any,
            "a".to_string(),
            Box::new(TaggedType::Concrete(Type::arrow(
                Type::arrow("a", "a"),
                Type::arrow("a", "a")
            )))
        )
    );

    let ast = parser
        .parse(Lexer::new("Any a => Any b => (a -> b) -> a -> b"))
        .unwrap();
    assert_eq!(
        ast,
        TaggedType::Tagged(
            TypeTag::Any,
            "a".to_string(),
            Box::new(TaggedType::Tagged(
                TypeTag::Any,
                "b".to_string(),
                Box::new(TaggedType::Concrete(Type::arrow(
                    Type::arrow("a", "b"),
                    Type::arrow("a", "b")
                )))
            ))
        )
    );

    let ast = parser.parse(Lexer::new("Nat -> Nat")).unwrap();
    assert_eq!(
        ast,
        TaggedType::Concrete(Type::arrow(PrimitiveType::Nat, PrimitiveType::Nat))
    );
}

#[test]
fn parse_abstraction() {
    let parser = parser::AstParser::new();
    let ast = parser.parse(Lexer::new(r"\x:Nat.x")).unwrap();
    assert_eq!(
        ast,
        Ast::Abstraction(
            "x".to_string(),
            Some(Type::Primitive(PrimitiveType::Nat).into()),
            Box::new(Ast::Variable("x".to_string()))
        )
    );

    let ast = parser.parse(Lexer::new(r"\x:Nat.\y:Nat.x")).unwrap();
    assert_eq!(
        ast,
        Ast::Abstraction(
            "x".to_string(),
            Some(Type::Primitive(PrimitiveType::Nat).into()),
            Box::new(Ast::Abstraction(
                "y".to_string(),
                Some(Type::Primitive(PrimitiveType::Nat).into()),
                Box::new(Ast::Variable("x".to_string()))
            ))
        )
    );

    let ast = parser
        .parse(Lexer::new(r"\x:Any a => a.\y:Any b => b.x"))
        .unwrap();
    assert_eq!(
        ast,
        Ast::Abstraction(
            "x".to_string(),
            Some(TaggedType::Tagged(
                TypeTag::Any,
                "a".to_string(),
                Box::new(TaggedType::Concrete("a".into()))
            )),
            Box::new(Ast::Abstraction(
                "y".to_string(),
                Some(TaggedType::Tagged(
                    TypeTag::Any,
                    "b".to_string(),
                    Box::new(TaggedType::Concrete("b".into()))
                )),
                Box::new(Ast::Variable("x".to_string()))
            ))
        )
    );

    let ast = parser.parse(Lexer::new(r"\x:Any a.\y:Any b.x")).unwrap();
    assert_eq!(
        ast,
        Ast::Abstraction(
            "x".to_string(),
            Some(TaggedType::Tagged(
                TypeTag::Any,
                "a".to_string(),
                Box::new(TaggedType::Concrete("a".into()))
            )),
            Box::new(Ast::Abstraction(
                "y".to_string(),
                Some(TaggedType::Tagged(
                    TypeTag::Any,
                    "b".to_string(),
                    Box::new(TaggedType::Concrete("b".into()))
                )),
                Box::new(Ast::Variable("x".to_string()))
            ))
        )
    );

    let ast = parser
        .parse(Lexer::new(r"\x:Any a => a,y:Any b => b.x"))
        .unwrap();
    assert_eq!(
        ast,
        Ast::Abstraction(
            "x".to_string(),
            Some(TaggedType::Tagged(
                TypeTag::Any,
                "a".to_string(),
                Box::new(TaggedType::Concrete("a".into()))
            )),
            Box::new(Ast::Abstraction(
                "y".to_string(),
                Some(TaggedType::Tagged(
                    TypeTag::Any,
                    "b".to_string(),
                    Box::new(TaggedType::Concrete("b".into()))
                )),
                Box::new(Ast::Variable("x".to_string()))
            ))
        )
    );

    let ast = parser.parse(Lexer::new(r"\x,y.x")).unwrap();
    assert_eq!(
        ast,
        Ast::Abstraction(
            "x".to_string(),
            None,
            Box::new(Ast::Abstraction(
                "y".to_string(),
                None,
                Box::new(Ast::Variable("x".to_string()))
            ))
        )
    );
}

#[test]
fn parse_application() {
    let parser = parser::AstParser::new();
    let ast = parser.parse(Lexer::new(r"(\x:Nat.x) 5")).unwrap();
    assert_eq!(
        ast,
        Ast::Application(
            Box::new(Ast::Abstraction(
                "x".to_string(),
                Some(Type::Primitive(PrimitiveType::Nat).into()),
                Box::new(Ast::Variable("x".to_string()))
            )),
            Box::new(Ast::Constant(Constant::Nat(5)))
        )
    );

    let ast = parser.parse(Lexer::new(r"add 1")).unwrap();
    assert_eq!(
        ast,
        Ast::Application(
            Box::new(Ast::Variable("add".to_string())),
            Box::new(Ast::Constant(Constant::Nat(1)))
        ),
    );

    let ast = parser.parse(Lexer::new(r"(add 1) 2")).unwrap();
    assert_eq!(
        ast,
        Ast::Application(
            Box::new(Ast::Application(
                Box::new(Ast::Variable("add".to_string())),
                Box::new(Ast::Constant(Constant::Nat(1)))
            )),
            Box::new(Ast::Constant(Constant::Nat(2)))
        ),
    );

    let ast = parser.parse(Lexer::new(r"add 1 2")).unwrap();
    assert_eq!(
        ast,
        Ast::Application(
            Box::new(Ast::Application(
                Box::new(Ast::Variable("add".to_string())),
                Box::new(Ast::Constant(Constant::Nat(1)))
            )),
            Box::new(Ast::Constant(Constant::Nat(2)))
        ),
    );

    let ast = parser.parse(Lexer::new(r"add 1 (x y)")).unwrap();
    assert_eq!(
        ast,
        Ast::Application(
            Box::new(Ast::Application(
                Box::new(Ast::Variable("add".to_string())),
                Box::new(Ast::Constant(Constant::Nat(1)))
            )),
            Box::new(Ast::Application(
                Box::new(Ast::Variable("x".to_string())),
                Box::new(Ast::Variable("y".to_string())),
            )),
        ),
    );

    let ast = parser
        .parse(Lexer::new(
            r"(\x:Any a => a -> a.\y:Num a => a. add 1 (x y)) (\x:Any a => a.x) 2",
        ))
        .unwrap();
    assert_eq!(
        ast,
        Ast::Application(
            Box::new(Ast::Application(
                Box::new(Ast::Abstraction(
                    "x".to_string(),
                    Some(TaggedType::Tagged(
                        TypeTag::Any,
                        "a".to_string(),
                        Box::new(TaggedType::Concrete(Type::arrow("a", "a")))
                    )),
                    Box::new(Ast::Abstraction(
                        "y".to_string(),
                        Some(TaggedType::Tagged(
                            TypeTag::Num,
                            "a".to_string(),
                            Box::new(TaggedType::Concrete("a".into()))
                        )),
                        Box::new(Ast::Application(
                            Box::new(Ast::Application(
                                Box::new(Ast::Variable("add".to_string())),
                                Box::new(Ast::Constant(Constant::Nat(1)))
                            )),
                            Box::new(Ast::Application(
                                Box::new(Ast::Variable("x".to_string())),
                                Box::new(Ast::Variable("y".to_string())),
                            ))
                        ))
                    )),
                )),
                Box::new(Ast::Abstraction(
                    "x".to_string(),
                    Some(TaggedType::Tagged(
                        TypeTag::Any,
                        "a".to_string(),
                        Box::new(TaggedType::Concrete("a".into()))
                    )),
                    Box::new(Ast::Variable("x".to_string()))
                ))
            )),
            Box::new(Ast::Constant(Constant::Nat(2)))
        )
    );
}

// (\x:Any a => a -> a.\y:Num a => a. add 1 (x y)) (\x:Any a => a.x) 2
