use std::{
    fmt::{self, Display, Formatter},
    num::{ParseFloatError, ParseIntError},
    str::ParseBoolError,
};

use logos::{Logos, SpannedIter};

use super::ParseError;

impl From<ParseIntError> for ParseError {
    fn from(err: ParseIntError) -> Self {
        Self::InvalidInteger(err)
    }
}

impl From<ParseFloatError> for ParseError {
    fn from(err: ParseFloatError) -> Self {
        Self::InvalidFloat(err)
    }
}

impl From<ParseBoolError> for ParseError {
    fn from(err: ParseBoolError) -> Self {
        Self::InvalidBool(err)
    }
}

#[derive(Logos, Debug, Clone)]
#[logos(skip r"[ \t\n\r\f]+", error = ParseError)]
pub enum Token {
    // #[regex(r#""([^\"\\]|\\.)*""#, |lex| {let slice = lex.slice(); slice[1..slice.len() - 1].to_string()})]
    // String(String),
    #[regex(r#"\d+"#, |lex| lex.slice().parse::<usize>())]
    Nat(usize),
    #[regex(r#"-?\d+\.\d+"#, |lex| lex.slice().parse::<f64>())]
    Float(f64),
    #[regex(r#"(true|false)"#, |lex| lex.slice().parse::<bool>())]
    Bool(bool),

    #[regex(r#"[_a-zA-Z][_a-zA-Z0-9]*'*"#, |lex| lex.slice().to_string())]
    Ident(String),

    #[token(r"\")]
    Lambda,
    #[token("=>")]
    FatArrow,
    #[token("->")]
    ThinArrow,
    #[token(":")]
    Colon,
    #[token(".")]
    Period,
    #[token(",")]
    Comma,

    #[token("=")]
    Equals,
    #[token("<>")]
    NotEquals,
    #[token("+")]
    Plus,
    #[token("-")]
    Minus,
    #[token("*")]
    Star,
    #[token("/")]
    Slash,
    #[token("^")]
    Hat,
    #[token("<")]
    Less,
    #[token(">")]
    Greater,
    #[token("<=")]
    LessEq,
    #[token(">=")]
    GreaterEq,
    #[token(r"/\")]
    Conjunction,
    #[token(r"\/")]
    Disjunction,

    #[token("(")]
    ParenOpen,
    #[token(")")]
    ParenClose,
}

pub(super) type Spanned<Tok, Loc, Error> = Result<(Loc, Tok, Loc), Error>;

pub(super) struct Lexer<'input> {
    // instead of an iterator over characters, we have a token iterator
    token_stream: SpannedIter<'input, Token>,
}
impl<'input> Lexer<'input> {
    pub fn new(input: &'input str) -> Self {
        // the Token::lexer() method is provided by the Logos trait
        Self {
            token_stream: Token::lexer(input).spanned(),
        }
    }
}
impl Iterator for Lexer<'_> {
    type Item = Spanned<Token, usize, ParseError>;

    fn next(&mut self) -> Option<Self::Item> {
        self.token_stream
            .next()
            .map(|(token, span)| Ok((span.start, token?, span.end)))
    }
}

impl Display for Token {
    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
        match self {
            Token::Lambda => write!(f, r"\"),
            Token::FatArrow => write!(f, "=>"),
            Token::ThinArrow => write!(f, "->"),
            Token::Colon => write!(f, ":"),
            Token::Period => write!(f, "."),
            Token::Comma => write!(f, ","),
            Token::Equals => write!(f, "="),
            Token::NotEquals => write!(f, "<>"),
            Token::Plus => write!(f, "+"),
            Token::Minus => write!(f, "-"),
            Token::Star => write!(f, "*"),
            Token::Slash => write!(f, "/"),
            Token::Hat => write!(f, "^"),
            Token::Less => write!(f, "<"),
            Token::Greater => write!(f, ">"),
            Token::LessEq => write!(f, "<="),
            Token::GreaterEq => write!(f, ">="),
            Token::Conjunction => write!(f, r"/\"),
            Token::Disjunction => write!(f, r"\/"),
            Token::ParenOpen => write!(f, "("),
            Token::ParenClose => write!(f, ")"),
            Token::Nat(n) => write!(f, "{n}"),
            Token::Float(n) => write!(f, "{n}"),
            Token::Bool(b) => write!(f, "{b}"),
            Token::Ident(i) => write!(f, "{i}"),
        }
    }
}
