use std::{collections::HashSet, fmt::Display};

use crate::parse::Ident;

use super::tagged::TypeTag;

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub enum PrimitiveType {
    Nat,
    Bool,
    Float,
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub enum Type {
    Variable(Ident),             // a
    Primitive(PrimitiveType),    // Bool | Nat
    Arrow(Box<Type>, Box<Type>), // 0 -> 1
}

impl From<PrimitiveType> for Type {
    fn from(value: PrimitiveType) -> Self {
        Self::Primitive(value)
    }
}

impl From<&str> for Type {
    fn from(value: &str) -> Self {
        Self::Variable(value.to_string())
    }
}

impl Type {
    pub fn arrow<T1: Into<Type>, T2: Into<Type>>(t1: T1, t2: T2) -> Self {
        Self::Arrow(Box::new(t1.into()), Box::new(t2.into()))
    }

    pub fn split_arrow(self) -> Option<(Type, Type)> {
        match self {
            Type::Arrow(lhs, rhs) => Some((*lhs, *rhs)),
            _ => None,
        }
    }

    pub fn is_concrete(&self) -> bool {
        match self {
            Type::Variable(_) => false,
            Type::Primitive(_) => true,
            Type::Arrow(t1, t2) => t1.is_concrete() && t2.is_concrete(),
        }
    }

    pub fn has_tag(&self, tag: &TypeTag) -> bool {
        match self {
            Type::Variable(_) => true,
            Type::Primitive(primitive_type) => matches!(
                (primitive_type, tag),
                (_, TypeTag::Any) | (PrimitiveType::Nat | PrimitiveType::Float, TypeTag::Num)
            ),
            Type::Arrow(_, _) => false,
        }
    }

    pub fn map_name<F: Fn(&mut String)>(&mut self, f: &F) {
        match self {
            Type::Variable(v) => f(v),
            Type::Arrow(lhs, rhs) => {
                lhs.map_name(f);
                rhs.map_name(f);
            }
            Type::Primitive(_) => {}
        }
    }

    pub fn name_used(&self, ident: &str) -> bool {
        match self {
            Type::Variable(i) if *i == ident => true,
            Type::Arrow(lhs, rhs) => lhs.name_used(ident) || rhs.name_used(ident),
            _ => false,
        }
    }

    pub fn type_var(&self) -> Option<String> {
        match self {
            Type::Variable(v) => Some(v.clone()),
            Type::Primitive(_) | Type::Arrow(_, _) => None,
        }
    }

    pub fn type_vars(&self) -> HashSet<String> {
        match self {
            Type::Variable(v) => {
                let mut set = HashSet::new();
                set.insert(v.to_string());
                set
            }
            Type::Primitive(_) => HashSet::new(),
            Type::Arrow(lhs, rhs) => {
                let mut vars = lhs.type_vars();
                for var in rhs.type_vars() {
                    vars.insert(var);
                }
                vars
            }
        }
    }

    pub fn specialize(self, ident: &str, typ: &Type) -> Type {
        match self {
            Type::Variable(i) if i == ident => typ.clone(),
            Type::Arrow(lhs, rhs) => Type::Arrow(
                Box::new(lhs.specialize(ident, typ)),
                Box::new(rhs.specialize(ident, typ)),
            ),
            t => t,
        }
    }
}

impl Display for Type {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Type::Variable(i) => write!(f, "{i}"),
            Type::Primitive(primitive_type) => write!(f, "{primitive_type}"),
            Type::Arrow(t1, t2) => write!(f, "({t1} -> {t2})"),
        }
    }
}

impl Display for PrimitiveType {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            PrimitiveType::Nat => write!(f, "Nat"),
            PrimitiveType::Bool => write!(f, "Bool"),
            PrimitiveType::Float => write!(f, "Float"),
        }
    }
}
